"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const socket_io_client_1 = require("socket.io-client");
const rxjs_1 = require("rxjs");
const operators_1 = require("rxjs/operators");
const util_1 = require("./util");
const ring_types_1 = require("./ring-types");
const rest_client_1 = require("./rest-client");
const deviceListMessageType = 'DeviceInfoDocGetList';
function flattenDeviceData(data) {
    return Object.assign({}, data.general && data.general.v2, data.device && data.device.v1);
}
class RingDevice {
    constructor(initialData, location, assetId) {
        this.initialData = initialData;
        this.location = location;
        this.assetId = assetId;
        this.onData = new rxjs_1.BehaviorSubject(this.initialData);
        this.zid = this.initialData.zid;
        this.id = this.zid;
        this.deviceType = this.initialData.deviceType;
        location.onDeviceDataUpdate
            .pipe(operators_1.filter(update => update.zid === this.zid))
            .subscribe(update => this.updateData(update));
    }
    updateData(update) {
        this.onData.next(Object.assign({}, this.data, update));
    }
    get data() {
        return this.onData.getValue();
    }
    get name() {
        return this.data.name;
    }
    get supportsVolume() {
        return (ring_types_1.deviceTypesWithVolume.includes(this.data.deviceType) &&
            this.data.volume !== undefined);
    }
    setVolume(volume) {
        if (isNaN(volume) || volume < 0 || volume > 1) {
            throw new Error('Volume must be between 0 and 1');
        }
        if (!this.supportsVolume) {
            throw new Error(`Volume can only be set on ${ring_types_1.deviceTypesWithVolume.join(', ')}`);
        }
        return this.setInfo({ device: { v1: { volume } } });
    }
    setInfo(body) {
        return this.location.sendMessage({
            msg: 'DeviceInfoSet',
            datatype: 'DeviceInfoSetType',
            dst: this.assetId,
            body: [
                Object.assign({ zid: this.zid }, body)
            ]
        });
    }
    sendCommand(commandType, data = {}) {
        this.setInfo({
            command: {
                v1: [
                    {
                        commandType,
                        data
                    }
                ]
            }
        });
    }
    toString() {
        return this.toJSON();
    }
    toJSON() {
        return JSON.stringify({
            data: this.data
        }, null, 2);
    }
}
exports.RingDevice = RingDevice;
class Location {
    constructor(locationDetails, cameras, hasHubs, restClient) {
        this.locationDetails = locationDetails;
        this.cameras = cameras;
        this.hasHubs = hasHubs;
        this.restClient = restClient;
        this.seq = 1;
        this.onMessage = new rxjs_1.Subject();
        this.onDataUpdate = new rxjs_1.Subject();
        this.onDeviceDataUpdate = this.onDataUpdate.pipe(operators_1.filter(message => {
            return message.datatype === 'DeviceInfoDocType' && Boolean(message.body);
        }), operators_1.concatMap(message => message.body), operators_1.map(flattenDeviceData));
        this.onDeviceList = this.onMessage.pipe(operators_1.filter(m => m.msg === deviceListMessageType));
        this.onDevices = this.onDeviceList.pipe(operators_1.scan((devices, { body: deviceList, src }) => {
            if (!deviceList) {
                return devices;
            }
            if (!this.receivedAssetDeviceLists.includes(src)) {
                this.receivedAssetDeviceLists.push(src);
            }
            return deviceList.reduce((updatedDevices, data) => {
                const flatData = flattenDeviceData(data), existingDevice = updatedDevices.find(x => x.zid === flatData.zid);
                if (existingDevice) {
                    existingDevice.updateData(flatData);
                    return updatedDevices;
                }
                return [...updatedDevices, new RingDevice(flatData, this, src)];
            }, devices);
        }, []), operators_1.distinctUntilChanged((a, b) => a.length === b.length), operators_1.filter(() => {
            return Boolean(this.assets &&
                this.assets.every(asset => this.receivedAssetDeviceLists.includes(asset.uuid)));
        }), operators_1.publishReplay(1), operators_1.refCount());
        this.onSessionInfo = this.onDataUpdate.pipe(operators_1.filter(m => m.msg === 'SessionInfo'), operators_1.map(m => m.body));
        this.onConnected = new rxjs_1.BehaviorSubject(false);
        this.reconnecting = false;
        this.receivedAssetDeviceLists = [];
        this.offlineAssets = [];
        this.locationId = locationDetails.location_id;
        // start listening for devices immediately
        this.onDevices.subscribe();
        // watch for sessions to come online
        this.onSessionInfo.subscribe(sessions => {
            sessions.forEach(({ connectionStatus, assetUuid }) => {
                const assetWasOffline = this.offlineAssets.includes(assetUuid);
                const asset = this.assets && this.assets.find(x => x.uuid === assetUuid);
                if (!asset) {
                    // we don't know about this asset, so don't worry about it
                    return;
                }
                if (connectionStatus === 'online') {
                    if (assetWasOffline) {
                        this.requestList(deviceListMessageType, assetUuid);
                        this.offlineAssets = this.offlineAssets.filter(id => id !== assetUuid);
                        util_1.logInfo(`Ring ${asset.kind} ${assetUuid} has come back online`);
                    }
                }
                else if (!assetWasOffline) {
                    util_1.logError(`Ring ${asset.kind} ${assetUuid} is offline or on cellular backup.  Waiting for status to change`);
                    this.offlineAssets.push(assetUuid);
                }
            });
        });
    }
    createConnection() {
        return __awaiter(this, void 0, void 0, function* () {
            util_1.logInfo('Creating location socket.io connection');
            const { assets, ticket, host } = yield this.restClient.request({
                url: 'https://app.ring.com/api/v1/clap/tickets?locationID=' + this.locationId
            });
            this.assets = assets;
            this.receivedAssetDeviceLists.length = 0;
            this.offlineAssets.length = 0;
            if (!assets.length) {
                const errorMessage = `No assets (alarm hubs or beam bridges) found for location ${this.locationDetails.name} - ${this.locationId}`;
                util_1.logError(errorMessage);
                throw new Error(errorMessage);
            }
            const connection = socket_io_client_1.connect(`wss://${host}/?authcode=${ticket}&ack=false&EIO=3`, { transports: ['websocket'] });
            const reconnect = () => {
                if (this.reconnecting && this.connectionPromise) {
                    return this.connectionPromise;
                }
                this.onConnected.next(false);
                util_1.logInfo('Reconnecting location socket.io connection');
                this.reconnecting = true;
                connection.close();
                return (this.connectionPromise = util_1.delay(1000).then(() => {
                    return this.createConnection();
                }));
            };
            this.reconnecting = false;
            connection.on('DataUpdate', (message) => {
                if (message.datatype === 'HubDisconnectionEventType') {
                    util_1.logInfo('Location connection told to reconnect');
                    return reconnect();
                }
                this.onDataUpdate.next(message);
            });
            connection.on('message', (message) => this.onMessage.next(message));
            connection.on('error', reconnect);
            connection.on('disconnect', reconnect);
            return new Promise((resolve, reject) => {
                connection.once('connect', () => {
                    resolve(connection);
                    this.onConnected.next(true);
                    util_1.logInfo('Ring connected to socket.io server');
                    assets.forEach(asset => this.requestList(deviceListMessageType, asset.uuid));
                });
                connection.once('error', reject);
            }).catch(reconnect);
        });
    }
    getConnection() {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.hasHubs) {
                throw new Error(`Location ${this.locationDetails.name} does not have any hubs`);
            }
            if (this.connectionPromise) {
                return this.connectionPromise;
            }
            return (this.connectionPromise = this.createConnection());
        });
    }
    sendMessage(message) {
        return __awaiter(this, void 0, void 0, function* () {
            const connection = yield this.getConnection();
            message.seq = this.seq++;
            connection.emit('message', message);
        });
    }
    sendCommandToSecurityPanel(commandType, data) {
        return __awaiter(this, void 0, void 0, function* () {
            const securityPanel = yield this.getSecurityPanel();
            return securityPanel.sendCommand(commandType, data);
        });
    }
    setAlarmMode(alarmMode, bypassSensorZids) {
        return this.sendCommandToSecurityPanel('security-panel.switch-mode', {
            mode: alarmMode,
            bypass: bypassSensorZids
        });
    }
    soundSiren() {
        return this.sendCommandToSecurityPanel('security-panel.sound-siren');
    }
    silenceSiren() {
        return this.sendCommandToSecurityPanel('security-panel.silence-siren');
    }
    setLightGroup(groupId, on, durationSeconds = 60) {
        this.restClient.request({
            method: 'POST',
            url: `https://api.ring.com/groups/v1/locations/${this.locationId}/groups/${groupId}/devices`,
            data: {
                lights_on: {
                    duration_seconds: durationSeconds,
                    enabled: on
                }
            },
            json: true
        });
    }
    getNextMessageOfType(type, src) {
        return this.onMessage
            .pipe(operators_1.filter(m => m.msg === type && m.src === src), operators_1.map(m => m.body), operators_1.take(1))
            .toPromise();
    }
    requestList(listType, assetId) {
        this.sendMessage({ msg: listType, dst: assetId });
    }
    getList(listType, assetId) {
        this.requestList(listType, assetId);
        return this.getNextMessageOfType(listType, assetId);
    }
    getDevices() {
        if (!this.hasHubs) {
            return [];
        }
        if (!this.connectionPromise) {
            this.getConnection();
        }
        return this.onDevices.pipe(operators_1.take(1)).toPromise();
    }
    getRoomList(assetId) {
        return this.getList('RoomGetList', assetId);
    }
    getSecurityPanel() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.securityPanel) {
                return this.securityPanel;
            }
            const devices = yield this.getDevices();
            const securityPanel = devices.find(device => {
                return device.data.deviceType === ring_types_1.RingDeviceType.SecurityPanel;
            });
            if (!securityPanel) {
                throw new Error(`Could not find a security panel for location ${this.locationDetails.name} - ${this.locationId}`);
            }
            return (this.securityPanel = securityPanel);
        });
    }
    disarm() {
        return this.setAlarmMode('none');
    }
    armHome(bypassSensorZids) {
        return this.setAlarmMode('some', bypassSensorZids);
    }
    armAway(bypassSensorZids) {
        return this.setAlarmMode('all', bypassSensorZids);
    }
    getEvents(limit = 50, olderThanId) {
        const paginationKey = olderThanId ? `&pagination_key=${olderThanId}` : '';
        return this.restClient.request({
            url: rest_client_1.clientApi(`locations/${this.locationId}/events?limit=${limit}${paginationKey}`)
        });
    }
}
exports.Location = Location;
