"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const axios_1 = require("axios");
const util_1 = require("./util");
const querystring = require("querystring");
const ringErrorCodes = {
    7050: 'NO_ASSET',
    7019: 'ASSET_OFFLINE',
    7061: 'ASSET_CELL_BACKUP',
    7062: 'UPDATING',
    7063: 'MAINTENANCE'
}, clientApiBaseUrl = 'https://api.ring.com/clients_api/', apiVersion = 11;
function clientApi(path) {
    return clientApiBaseUrl + path;
}
exports.clientApi = clientApi;
function requestWithRetry(options) {
    return __awaiter(this, void 0, void 0, function* () {
        try {
            util_1.logInfo(`Making request: ${JSON.stringify(options, null, 2)}`);
            const { data, headers } = yield axios_1.default(options);
            if (typeof data === 'object' && headers.date) {
                data.responseTimestamp = new Date(headers.date).getTime();
            }
            return data;
        }
        catch (e) {
            if (!e.response) {
                util_1.logError(`Failed to reach Ring server at ${options.url}.  Trying again in 5 seconds...`);
                yield util_1.delay(5000);
                return requestWithRetry(options);
            }
            throw e;
        }
    });
}
class RingRestClient {
    constructor(authOptions) {
        this.authOptions = authOptions;
        // prettier-ignore
        this.refreshToken = ('refreshToken' in this.authOptions ? this.authOptions.refreshToken : undefined);
        this.authPromise = this.getAuthToken();
        this.sessionPromise = this.getSession();
        this.using2fa = false;
    }
    getGrantData(twoFactorAuthCode) {
        if (this.refreshToken && !twoFactorAuthCode) {
            return {
                grant_type: 'refresh_token',
                refresh_token: this.refreshToken
            };
        }
        const { authOptions } = this;
        if ('email' in authOptions) {
            return {
                grant_type: 'password',
                password: authOptions.password,
                username: authOptions.email
            };
        }
        throw new Error('Refresh token is not valid.  Unable to authenticate with Ring servers.');
    }
    getAuthToken(twoFactorAuthCode) {
        return __awaiter(this, void 0, void 0, function* () {
            const grantData = this.getGrantData(twoFactorAuthCode), twoFactorAuthHeaders = twoFactorAuthCode
                ? {
                    '2fa-code': twoFactorAuthCode,
                    '2fa-Support': 'true'
                }
                : {};
            try {
                const response = yield requestWithRetry({
                    url: 'https://oauth.ring.com/oauth/token',
                    data: Object.assign({ client_id: 'ring_official_android', scope: 'client' }, grantData),
                    method: 'POST',
                    headers: Object.assign({ 'content-type': 'application/json' }, twoFactorAuthHeaders)
                });
                this.refreshToken = response.refresh_token;
                return response;
            }
            catch (requestError) {
                if (grantData.refresh_token) {
                    // failed request with refresh token, try again with username/password
                    this.refreshToken = undefined;
                    return this.getAuthToken();
                }
                const response = requestError.response || {}, responseData = response.data || {}, responseError = typeof responseData.error === 'string' ? responseData.error : '';
                if (response.status === 412 || // need 2fa code
                    (response.status === 400 &&
                        responseError.startsWith('Verification Code')) // invalid 2fa code entered
                ) {
                    const code = yield util_1.requestInput('Ring 2fa enabled.  Please enter code from text message: ');
                    this.using2fa = true;
                    return this.getAuthToken(code);
                }
                const authTypeMessage = 'refreshToken' in this.authOptions
                    ? 'refresh token is'
                    : 'email and password are', errorMessage = `Failed to fetch oauth token from Ring. Verify that your ${authTypeMessage} correct. ` +
                    responseError;
                util_1.logError(requestError.response);
                util_1.logError(errorMessage);
                throw new Error(errorMessage);
            }
        });
    }
    fetchNewSession(authToken) {
        return __awaiter(this, void 0, void 0, function* () {
            const hardwareId = util_1.generateRandomId();
            yield requestWithRetry({
                url: clientApi('session'),
                data: {
                    device: {
                        hardware_id: hardwareId,
                        metadata: {
                            api_version: apiVersion
                        },
                        os: 'android'
                    }
                },
                method: 'POST',
                headers: {
                    authorization: `Bearer ${authToken.access_token}`,
                    'content-type': 'application/json'
                }
            });
            return { hardwareId };
        });
    }
    getSession() {
        return this.authPromise.then((authToken) => __awaiter(this, void 0, void 0, function* () {
            try {
                return yield this.fetchNewSession(authToken);
            }
            catch (e) {
                if (e && e.response && e.response.status === 429) {
                    const retryAfter = e.response.headers['retry-after'], waitSeconds = isNaN(retryAfter) ? 200 : Number.parseInt(retryAfter);
                    util_1.logError(`Session response rate limited. Waiting to retry for ${waitSeconds} seconds`);
                    yield util_1.delay((waitSeconds + 1) * 1000);
                    util_1.logInfo('Retrying session request');
                    return this.getSession();
                }
                throw e;
            }
        }));
    }
    refreshAuth() {
        this.authPromise = this.getAuthToken();
        this.refreshSession();
    }
    refreshSession() {
        this.sessionPromise = this.getSession();
    }
    request(options) {
        return __awaiter(this, void 0, void 0, function* () {
            const { method, url, data, json, responseType } = options, authTokenResponse = yield this.authPromise, session = yield this.sessionPromise, headers = {
                'content-type': json
                    ? 'application/json'
                    : 'application/x-www-form-urlencoded',
                authorization: `Bearer ${authTokenResponse.access_token}`,
                hardware_id: session.hardwareId
            };
            try {
                return yield requestWithRetry({
                    method: method || 'GET',
                    url,
                    data: json ? data : querystring.stringify(data),
                    headers,
                    responseType
                });
            }
            catch (e) {
                const response = e.response || {};
                if (response.status === 401) {
                    this.refreshAuth();
                    return this.request(options);
                }
                if (response.status === 404 &&
                    response.data &&
                    Array.isArray(response.data.errors)) {
                    const errors = response.data.errors, errorText = errors
                        .map((code) => ringErrorCodes[code])
                        .filter((x) => x)
                        .join(', ');
                    if (errorText) {
                        util_1.logError(`http request failed.  ${url} returned errors: (${errorText}).  Trying again in 20 seconds`);
                        yield util_1.delay(20000);
                        return this.request(options);
                    }
                    else {
                        util_1.logError(`http request failed.  ${url} returned unknown errors: (${errors}).`);
                    }
                }
                if (response.status === 404 && url.startsWith(clientApiBaseUrl)) {
                    util_1.logError('Session hardware_id not found.  Creating a new session and trying again.');
                    this.refreshSession();
                    return this.request(options);
                }
                util_1.logError(`Request to ${url} failed`);
                throw e;
            }
        });
    }
}
exports.RingRestClient = RingRestClient;
