"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const ring_types_1 = require("./ring-types");
const rest_client_1 = require("./rest-client");
const rxjs_1 = require("rxjs");
const operators_1 = require("rxjs/operators");
const util_1 = require("./util");
const snapshotTimestampGracePeriod = 10000;
function getBatteryLevel(data) {
    const batteryLevel = typeof data.battery_life === 'number'
        ? data.battery_life
        : Number.parseFloat(data.battery_life);
    if (isNaN(batteryLevel)) {
        return null;
    }
    return batteryLevel;
}
class RingCamera {
    constructor(initialData, isDoorbot, restClient) {
        this.initialData = initialData;
        this.isDoorbot = isDoorbot;
        this.restClient = restClient;
        this.id = this.initialData.id;
        this.deviceType = this.initialData.kind;
        this.model = ring_types_1.RingCameraModel[this.initialData.kind] || 'Unknown Model';
        this.hasLight = this.initialData.led_status !== undefined;
        this.hasSiren = this.initialData.siren_status !== undefined;
        this.onData = new rxjs_1.BehaviorSubject(this.initialData);
        this.onRequestUpdate = new rxjs_1.Subject();
        this.onNewDing = new rxjs_1.Subject();
        this.onActiveDings = new rxjs_1.BehaviorSubject([]);
        this.onDoorbellPressed = this.onNewDing.pipe(operators_1.filter(ding => ding.kind === 'ding'), operators_1.share());
        this.onMotionDetected = this.onActiveDings.pipe(operators_1.map(dings => dings.some(ding => ding.motion || ding.kind === 'motion')), operators_1.distinctUntilChanged(), operators_1.publishReplay(1), operators_1.refCount());
        this.onBatteryLevel = this.onData.pipe(operators_1.map(getBatteryLevel), operators_1.distinctUntilChanged());
        this.hasSlowSnapshotRefresh = this.deviceType === 'doorbell_v3'; // only refreshes timestamp every 10 minutes
    }
    updateData(update) {
        this.onData.next(update);
    }
    requestUpdate() {
        this.onRequestUpdate.next();
    }
    get data() {
        return this.onData.getValue();
    }
    get name() {
        return this.data.description;
    }
    get activeDings() {
        return this.onActiveDings.getValue();
    }
    get batteryLevel() {
        return getBatteryLevel(this.data);
    }
    doorbotUrl(path) {
        return rest_client_1.clientApi(`doorbots/${this.id}/${path}`);
    }
    setLight(on) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.hasLight) {
                return false;
            }
            const state = on ? 'on' : 'off';
            yield this.restClient.request({
                method: 'PUT',
                url: this.doorbotUrl('floodlight_light_' + state)
            });
            this.updateData(Object.assign({}, this.data, { led_status: state }));
            return true;
        });
    }
    setSiren(on) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.hasSiren) {
                return false;
            }
            const state = on ? 'on' : 'off';
            yield this.restClient.request({
                method: 'PUT',
                url: this.doorbotUrl('siren_' + state)
            });
            this.updateData(Object.assign({}, this.data, { siren_status: { seconds_remaining: 1 } }));
            return true;
        });
    }
    getHealth() {
        return __awaiter(this, void 0, void 0, function* () {
            const response = yield this.restClient.request({
                url: this.doorbotUrl('health')
            });
            return response.device_health;
        });
    }
    startVideoOnDemand() {
        return this.restClient.request({
            method: 'POST',
            url: this.doorbotUrl('vod')
        });
    }
    getSipConnectionDetails() {
        return __awaiter(this, void 0, void 0, function* () {
            const vodPromise = this.onNewDing
                .pipe(operators_1.filter(x => x.kind === 'on_demand'), operators_1.take(1))
                .toPromise();
            yield this.startVideoOnDemand();
            return vodPromise;
        });
    }
    processActiveDing(ding) {
        const activeDings = this.activeDings;
        this.onNewDing.next(ding);
        this.onActiveDings.next(activeDings.concat([ding]));
        setTimeout(() => {
            const allActiveDings = this.activeDings, otherDings = allActiveDings.filter(oldDing => oldDing !== ding);
            this.onActiveDings.next(otherDings);
        }, 65 * 1000); // dings last ~1 minute
    }
    getHistory(limit = 10, favoritesOnly = false) {
        const favoritesParam = favoritesOnly ? '&favorites=1' : '';
        return this.restClient.request({
            url: this.doorbotUrl(`history?limit=${limit}${favoritesParam}`)
        });
    }
    getRecording(dingIdStr) {
        return __awaiter(this, void 0, void 0, function* () {
            const response = yield this.restClient.request({
                url: rest_client_1.clientApi(`dings/${dingIdStr}/share/play?disable_redirect=true`)
            });
            return response.url;
        });
    }
    updateTimestamp() {
        return __awaiter(this, void 0, void 0, function* () {
            const { timestamps, responseTimestamp } = yield this.restClient.request({
                url: rest_client_1.clientApi(`snapshots/timestamps`),
                method: 'POST',
                data: {
                    doorbot_ids: [this.id]
                },
                json: true
            }), timestamp = timestamps[0];
            return {
                timestamp: timestamp ? timestamp.timestamp : 0,
                responseTimestamp
            };
        });
    }
    refreshSnapshot(allowStale) {
        return __awaiter(this, void 0, void 0, function* () {
            const slowSnapshots = this.hasSlowSnapshotRefresh && !allowStale, snapshotRefreshDelay = slowSnapshots ? 2000 : 500, maxSnapshotRefreshSeconds = slowSnapshots
                ? 600 // 10 minutes if waiting for slow snapshots
                : this.hasSlowSnapshotRefresh
                    ? 5 // fail quickly for slow snapshot devices when stale images are ok
                    : 30, // wait 30 seconds for normal devices
            maxSnapshotRefreshAttempts = (maxSnapshotRefreshSeconds * 1000) / snapshotRefreshDelay;
            for (let i = 0; i < maxSnapshotRefreshAttempts; i++) {
                const { timestamp, responseTimestamp } = yield this.updateTimestamp(), timestampAge = Math.abs(responseTimestamp - timestamp);
                if (timestampAge < snapshotTimestampGracePeriod) {
                    return;
                }
                yield util_1.delay(snapshotRefreshDelay);
            }
            throw new Error(`Snapshot failed to refresh after ${maxSnapshotRefreshAttempts} attempts`);
        });
    }
    getSnapshot(allowStale = false) {
        return __awaiter(this, void 0, void 0, function* () {
            this.refreshSnapshotInProgress =
                this.refreshSnapshotInProgress || this.refreshSnapshot(allowStale);
            try {
                yield this.refreshSnapshotInProgress;
            }
            catch (e) {
                util_1.logError(e);
            }
            this.refreshSnapshotInProgress = undefined;
            return this.restClient.request({
                url: rest_client_1.clientApi(`snapshots/image/${this.id}`),
                responseType: 'arraybuffer'
            });
        });
    }
}
exports.RingCamera = RingCamera;
