"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const api_1 = require("../api");
const hap_1 = require("./hap");
const security_panel_1 = require("./security-panel");
const base_station_1 = require("./base-station");
const keypad_1 = require("./keypad");
const contact_sensor_1 = require("./contact-sensor");
const motion_sensor_1 = require("./motion-sensor");
const lock_1 = require("./lock");
const smoke_alarm_1 = require("./smoke-alarm");
const co_alarm_1 = require("./co-alarm");
const smoke_co_listener_1 = require("./smoke-co-listener");
const beam_1 = require("./beam");
const multi_level_switch_1 = require("./multi-level-switch");
const camera_1 = require("./camera");
const pluginName = 'homebridge-ring-alarm', platformName = 'RingAlarm', debug = __filename.includes('release-homebridge');
process.env.RING_DEBUG = debug ? 'true' : '';
function getAccessoryClass(device) {
    const { deviceType } = device;
    switch (deviceType) {
        case api_1.RingDeviceType.ContactSensor:
            return contact_sensor_1.ContactSensor;
        case api_1.RingDeviceType.MotionSensor:
            return motion_sensor_1.MotionSensor;
        case api_1.RingDeviceType.SecurityPanel:
            return security_panel_1.SecurityPanel;
        case api_1.RingDeviceType.BaseStation:
            return base_station_1.BaseStation;
        case api_1.RingDeviceType.Keypad:
            return keypad_1.Keypad;
        case api_1.RingDeviceType.SmokeAlarm:
            return smoke_alarm_1.SmokeAlarm;
        case api_1.RingDeviceType.CoAlarm:
            return co_alarm_1.CoAlarm;
        case api_1.RingDeviceType.SmokeCoListener:
            return smoke_co_listener_1.SmokeCoListener;
        case api_1.RingDeviceType.BeamsMotionSensor:
        case api_1.RingDeviceType.BeamsSwitch:
        case api_1.RingDeviceType.BeamsTransformerSwitch:
        case api_1.RingDeviceType.BeamsLightGroupSwitch:
            return beam_1.Beam;
        case api_1.RingDeviceType.MultiLevelSwitch:
            return multi_level_switch_1.MultiLevelSwitch;
    }
    if (/^lock($|\.)/.test(deviceType)) {
        return lock_1.Lock;
    }
    return null;
}
class RingAlarmPlatform {
    constructor(log, config, api) {
        this.log = log;
        this.config = config;
        this.api = api;
        this.homebridgeAccessories = {};
        if (!config) {
            this.log.info('No configuration found for platform RingAlarm');
            return;
        }
        config.cameraStatusPollingSeconds = config.cameraStatusPollingSeconds || 20;
        config.cameraDingsPollingSeconds = config.cameraDingsPollingSeconds || 1;
        this.api.on('didFinishLaunching', () => {
            this.log.debug('didFinishLaunching');
            this.connectToApi().catch(e => {
                this.log.error('Error connecting to API');
                this.log.error(e);
            });
        });
        this.homebridgeAccessories = {};
    }
    configureAccessory(accessory) {
        this.log.info(`Configuring cached accessory ${accessory.UUID} ${accessory.displayName}`);
        this.log.debug('%j', accessory);
        this.homebridgeAccessories[accessory.UUID] = accessory;
    }
    connectToApi() {
        return __awaiter(this, void 0, void 0, function* () {
            const ringApi = new api_1.RingApi(this.config), locations = yield ringApi.getLocations(), { api } = this, cachedAccessoryIds = Object.keys(this.homebridgeAccessories), platformAccessories = [], cameraAccessories = [], activeAccessoryIds = [];
            yield Promise.all(locations.map((location) => __awaiter(this, void 0, void 0, function* () {
                const devices = yield location.getDevices(), cameras = location.cameras, allDevices = [...devices, ...cameras];
                this.log.info(`Configuring ${cameras.length} cameras and ${devices.length} devices for locationId ${location.locationId}`);
                allDevices.forEach(device => {
                    const isCamera = device instanceof api_1.RingCamera, AccessoryClass = isCamera ? camera_1.Camera : getAccessoryClass(device), debugPrefix = debug ? 'TEST ' : '', cameraIdDifferentiator = isCamera ? 'camera' : '', // this forces bridged cameras from old version of the plugin to be seen as "stale"
                    id = debugPrefix + device.id.toString() + cameraIdDifferentiator, uuid = hap_1.hap.UUIDGen.generate(id);
                    if (!AccessoryClass ||
                        (this.config.hideLightGroups &&
                            device.deviceType === api_1.RingDeviceType.BeamsLightGroupSwitch)) {
                        return;
                    }
                    const createHomebridgeAccessory = () => {
                        const accessory = new hap_1.hap.PlatformAccessory(debugPrefix + device.name, uuid, isCamera
                            ? hap_1.hap.AccessoryCategories.CAMERA
                            : hap_1.hap.AccessoryCategories.SECURITY_SYSTEM);
                        this.log.info(`Adding new accessory ${device.deviceType} ${debugPrefix +
                            device.name}`);
                        if (isCamera) {
                            cameraAccessories.push(accessory);
                        }
                        else {
                            platformAccessories.push(accessory);
                        }
                        return accessory;
                    }, homebridgeAccessory = this.homebridgeAccessories[uuid] || createHomebridgeAccessory();
                    new AccessoryClass(device, homebridgeAccessory, this.log, this.config);
                    this.homebridgeAccessories[uuid] = homebridgeAccessory;
                    activeAccessoryIds.push(uuid);
                });
            })));
            if (platformAccessories.length) {
                api.registerPlatformAccessories(pluginName, platformName, platformAccessories);
            }
            if (cameraAccessories.length) {
                api.publishCameraAccessories(pluginName, cameraAccessories);
            }
            const staleAccessories = cachedAccessoryIds
                .filter(cachedId => !activeAccessoryIds.includes(cachedId))
                .map(id => this.homebridgeAccessories[id]);
            staleAccessories.forEach(staleAccessory => {
                this.log.info(`Removing stale cached accessory ${staleAccessory.UUID} ${staleAccessory.displayName}`);
            });
            if (staleAccessories.length) {
                this.api.unregisterPlatformAccessories(pluginName, platformName, staleAccessories);
            }
        });
    }
}
exports.RingAlarmPlatform = RingAlarmPlatform;
